#workflow_tree {
    width: 30;
    height: 100%;
    background: $surface;
    border-right: solid $primary;
    hatch: right $foreground 10%;
    padding: 1;
}

#node_details {
    height: 30%;
    background: $surface;
    hatch: right $foreground 10%;
    padding: 0;
    /*border-bottom: solid $primary;*/
}

#playbook_stdout {
    background: $surface-darken-1;
    height: 70%;
    padding: 1;
}

#action_buttons {
    display: none;
    padding: 0;
    width: 100%;
    align-horizontal: right;
}

#action_buttons Button {
    width: auto;
    height: auto;
    margin-left: 2;
}

Button {
    /* background: $primary;
    color: $text-primary; */
    border: none;
    min-width: 15;
}

Header {
    background: $primary-darken-1;
    color: $text-primary;
    text-style: bold;
}

Footer {
    background: $primary-darken-1;
    color: $text-primary;
}


Tree {
    padding: 1;
}

DataTable {
    padding: 1;
}

RichLog {
    padding: 0;
}


/* Styles for the quit confirmation dialog */
StopWorkflowScreen {
    align: center middle;
}

StopWorkflowScreen > Vertical {
    width: auto;
    height: auto;
    border: thick $background 80%;
    background: $surface;
}

StopWorkflowScreen > Vertical > Label,
StopWorkflowScreen > Vertical > DataTable {
    width: 100%;
    content-align-horizontal: center;
    margin-top: 1;
}

StopWorkflowScreen > Vertical > Horizontal {
    width: auto;
    height: auto;
}

StopWorkflowScreen > Vertical > Horizontal > Button {
    margin: 2 4;
}

/* QuitScreen dialog */
QuitScreen, DoubtfulNodeScreen{
    align: center middle;
}

DoubtfulNodeScreen > Container, QuitScreen > Container {
    width: auto;
    height: auto;
    border: thick $background 80%;
    background: $surface;
}

DoubtfulNodeScreen > Container > Label, QuitScreen > Container > Label {
    width: 100%;
    content-align-horizontal: center;
    margin-top: 1;
}

DoubtfulNodeScreen > Container > Horizontal, QuitScreen > Container > Horizontal {
    width: auto;
    height: auto;
}

DoubtfulNodeScreen > Container > Horizontal > Button, QuitScreen > Container > Horizontal > Button {
    margin: 2 4;
}

/* #dialog {
    width: 60;
    height: 10;
    border: thick $primary-lighten-2;
    background: $surface;
    padding: 1 2;
    align: center middle;
}

#question {
    width: 100%;
    text-align: center;
    margin-bottom: 1;
}

#buttons {
    width: 100%;
    align-horizontal: center;
    margin-top: 1;
} */
